/* aewm - a minimalist X11 window mananager. vim:sw=4:ts=4:et
 *
 * Copyright 1998-2005 Decklin Foster <decklin@red-bean.com>. This
 * program is free software; please see LICENSE for details. */

#include <stdlib.h>
#include <string.h>
#include "atom.h"

static char *get_string_atom(Window, Atom, Atom);

/* Get the window-manager name (aka human-readable "title") for a given
 * window. There are two ways a client can set this:
 *
 * 1. _NET_WM_STRING, which has type UTF8_STRING. This is preferred and
 *    is always used if available.
 * 2. WM_NAME, which has type COMPOUND_STRING or STRING. This is the old
 *    ICCCM way, which we fall back to in the absence of _NET_WM_STRING.
 *    In this case, we ask X to convert the value of the property to
 *    UTF-8 for us. N.b.: STRING is Latin-1 whatever the locale.
 *    COMPOUND_STRING is the most hideous abomination ever created.
 *    Thankfully we do not have to worry about any of this.
 *
 * If UTF-8 conversion is not available (XFree86 < 4.0.2, or any older X
 * implementation), only WM_NAME will be checked, and, at least for
 * XFree86 and X.Org, it will only be returned if it has type STRING.
 * This is due to an inherent limitation in their implementation of
 * XFetchName. If you have a different X vendor, YMMV.
 *
 * In all cases, this function asks X to allocate the returned string,
 * so it must be freed with XFree. */

char *get_wm_name(Window w)
{
    char *name;
#ifdef X_HAVE_UTF8_STRING
    XTextProperty name_prop;
    XTextProperty name_prop_converted;
    char **name_list;
    int nitems;

    if ((name = get_string_atom(w, net_wm_name, utf8_string))) {
        return name;
    } else if (XGetWMName(dpy, w, &name_prop)) {
        if (Xutf8TextPropertyToTextList(dpy, &name_prop, &name_list,
                &nitems) == Success && nitems >= 1) {
            /* Now we've got a freshly allocated XTextList. Since it
             * might have multiple items that need to be joined, and we
             * need to return something that can be freed by XFree, we
             * roll it back up into an XTextProperty. */
            if (Xutf8TextListToTextProperty(dpy, name_list, nitems,
                    XUTF8StringStyle, &name_prop_converted) == Success) {
                XFreeStringList(name_list);
                return name_prop_converted.value;
            } else {
                /* Not much we can do here. This should never happen
                 * anyway. */
                XFreeStringList(name_list);
                return NULL;
            }
        } else {
            return name_prop.value;
        }
    } else {
        /* There actually is no prop. There is only NULL! */
        return NULL;
    }
#else
    XFetchName(dpy, w, &name);
    return name;
#endif
}

/* I give up on trying to do this the right way. We'll just request as
 * many elements as possible. If that's not the entire string, we're
 * fucked. In reality this should never happen. (That's the second time
 * I get to say "this should never happen" in this file!)
 *
 * Standard gripe about casting nonsense applies. */

static char *get_string_atom(Window w, Atom a, Atom type)
{
    Atom real_type;
    int real_format = 0;
    unsigned long items_read = 0;
    unsigned long bytes_left = 0;
    unsigned char *data;

    XGetWindowProperty(dpy, w, a, 0, LONG_MAX, False, type,
        &real_type, &real_format, &items_read, &bytes_left, &data);

    /* XXX: should check bytes_left here and bail if nonzero. You know,
     * in case I wanted to store a >4gb string on the server.
     *
     * What, stop looking at me funny. */

    if (real_format == 8 && items_read >= 1)
        return (char *)data;
    else
        return NULL;
}

/* Somebody at central casting needs to be fired.
 *
 * In this case we just plain do not care about bytes_left, as the
 * purpose of this function is to grab a single, cardinal (not
 * necessarily CARDINAL... WINDOW, ATOM, whatever, anything that fits in
 * a long) value, whether it is the only value for that property or the
 * first of an array of values. If we needed arrays for some reason this
 * would obviously need an **nitems_ret or somesuch and a policy for
 * what ridiculous wrapper function (I'm looking at you, Xlib and GTK+)
 * to insist on freeing the data with. */

int get_atom(Window w, Atom a, Atom type, long *data_return)
{
    Atom real_type;
    int real_format = 0;
    unsigned long items_read = 0;
    unsigned long bytes_left = 0;
    unsigned char *data;

    XGetWindowProperty(dpy, w, a, 0, 1, False, type,
        &real_type, &real_format, &items_read, &bytes_left, &data);

    if (real_format == 32 && items_read >= 1) {
        *data_return = *(long *)data;
        XFree(data);
        return 1;
    } else {
        return 0;
    }
}

int set_atom(Window w, Atom a, Atom type, long val)
{
    return (XChangeProperty(dpy, w, a, type,
        32, PropModeReplace, (unsigned char *)&val, 1) == Success);
}

int append_to_atom(Window w, Atom a, Atom type, long new_val)
{
    return (XChangeProperty(dpy, w, a, type, 32,
        PropModePrepend, (unsigned char *)&new_val, 1) == Success);
}

void atom_foreach(Window w, Atom a, Atom type, atom_cb_func atom_cb,
    void *cb_data)
{
    Atom real_type;
    int real_format = 0;
    unsigned long items_read = 0;
    unsigned long bytes_left = 0;
    unsigned char *data;
    long start = 0;

    do {
        XGetWindowProperty(dpy, w, a, start, 1, False, type,
            &real_type, &real_format, &items_read, &bytes_left, &data);
        if (real_format == 32 && items_read >= 1) {
            atom_cb((long *)data, cb_data);
            XFree(data);
            start += items_read;
        }
    } while (real_format == 32 && bytes_left);
}
